from src.cli import console
from src.local.providers.abstract_provider import AbstractK8sProvider


def get_cluster_or_exit(ctx, project_id) -> AbstractK8sProvider:
    cluster_data = ctx.cluster_manager.get(id=project_id)
    cluster = ctx.cluster_manager.select(cluster_data=cluster_data)
    if not cluster:
        console.error("The project cluster does not exist.", _exit=True)

    return cluster



from unittest.mock import patch

import pytest
from click.testing import CliRunner

from src.cli import app
from unikube import ClickContext


def check():
    """Function used to mock check function"""
    pass


def test_list():
    runner = CliRunner()
    obj = ClickContext()
    obj.auth.check = check
    result = runner.invoke(
        app.list,
        obj=obj,
    )
    assert "[INFO] No project is running.\n" in result.output


def test_shell_cluster_not_found():
    runner = CliRunner()
    obj = ClickContext()
    obj.auth.check = check
    result = runner.invoke(
        app.shell,
        [
            "test",
            "--organization",
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
            "--project",
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
            "--deck",
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
        ],
        obj=obj,
    )
    assert "[ERROR] The project cluster could not be found or you have another project activated.\n" in result.output



import os
import sys

import click
from tabulate import tabulate

import src.cli.console as console
from src import settings
from src.local.dependency import install_dependency, probe_dependencies


@click.command()
@click.option("--reinstall", help="Reinstall the given dependencies")
def install(reinstall):
    """
    Install dependencies on your local machine.
    """

    def _do_install():
        incomplete = []
        successful = []
        unsuccessful = []
        for dependency in dependencies:
            rcode = install_dependency(dependency["name"])
            # since this run can contain multiple installations, capture all return codes
            if rcode is None:
                incomplete.append(dependency["name"])
            elif rcode == 0:
                successful.append(dependency["name"])
            elif rcode != 0:
                unsuccessful.append(dependency["name"])
        if unsuccessful:
            console.error("Some of the requested installations terminated unsuccessful")
        elif successful and not unsuccessful and not incomplete:
            # this only become 0 if installation actually run and was successful
            console.success("All requested dependencies installed successfully")
        elif incomplete:
            console.warning("Not all dependencies could be installed")

    # check account permission
    if os.geteuid() != 0:
        console.warning(
            "You are not running the installation with an administrative account. "
            "You may be prompted for your password."
        )

    # install
    if reinstall:
        dependencies = [{"name": i} for i in reinstall.split(",")]
    else:
        report_data = probe_dependencies(silent=True)
        dependencies = list(filter(lambda x: not x["success"], report_data))
        if len(dependencies) == 1:
            console.info(f"The following dependency is going to be installed: {dependencies[0]['name']}")
        elif len(dependencies) > 1:
            console.info(
                f"The following dependencies are going to be " f"installed: {','.join(k['name'] for k in dependencies)}"
            )
        else:
            console.info("All dependencies are already satisfied. No action taken.")
            sys.exit(0)

    _do_install()


@click.command()
@click.option("--verbose", "-v", is_flag=True, default=False, help="")
def verify(verbose):
    """
    Verifies the installation of dependencies on your local machine.
    """

    report_data = probe_dependencies(silent=verbose)
    unsuccessful = list(filter(lambda x: not x["success"], report_data))

    # show detailed table
    if verbose:
        successful = list(filter(lambda x: x["success"], report_data))

        console.table(
            successful + unsuccessful,
            headers={
                "name": "Name",
                "success": "Ok",
                "required_version": "Required Version",
                "installed_version": "Installed Version",
                "msg": "Message",
            },
        )

    if unsuccessful:
        console.error(
            f"There {'is' if len(unsuccessful) == 1 else 'are'} {len(unsuccessful)} (of {len(report_data)}) "
            f"unsuccessfully probed {'dependency' if len(unsuccessful) == 1 else 'dependencies'} on your "
            f"local machine. Please run 'unikube system install' in order to fix "
            f"these issues."
        )
        return False

    console.success("Local dependencies verified.")

    return True



import re
import sys
from urllib.parse import urljoin

import click_spinner
import requests
from requests import HTTPError, Session

import src.cli.console as console
from src import settings
from src.authentication.authentication import TokenAuthentication
from src.context import ClickContext
from src.graphql import EnvironmentType
from src.local.providers.types import K8sProviderType
from src.local.system import Telepresence


def select_entity(entity_list, identifier):
    # parsing id, which should be in parentheses after the project title
    id = re.search("(?<=\\()[^)]*(?=\\))", identifier)
    similar_entities = []
    if id:
        identifier = id.group(0)
    for entity in entity_list:
        if identifier == entity.get("id"):
            return entity
        elif identifier in (entity.get("title"), entity.get("slug")):
            similar_entities.append(entity)
    if len(similar_entities) > 1:
        console.warning(
            f"Entity {similar_entities[0].get('title')} has a duplicate title or slug. Specify ID directly "
            f"after title or slug in parentheses."
        )
        return None
    elif len(similar_entities) == 1:
        return similar_entities[0]
    else:
        console.warning(f"Entity {identifier} was not found.")
        return None


def select_entity_from_cluster_list(cluster_list, identifier):
    # parsing id, which should be in parentheses after the project title
    id = re.search("(?<=\\()[^)]*(?=\\))", identifier)
    similar_entities = []
    if id:
        identifier = id.group(0)
    for entity in cluster_list:
        if identifier == entity.id:
            return entity
        elif (hasattr(entity, "slug") and identifier == entity.slug) or (identifier == entity.name):
            similar_entities.append(entity)
    if len(similar_entities) > 1:
        console.warning(
            f"Entity {similar_entities[0].name} has a duplicate title or slug. Specify ID directly "
            f"after title or slug in parentheses."
        )
        return None
    elif len(similar_entities) == 1:
        return similar_entities[0]
    else:
        console.warning(f"Entity {identifier} was not found.")
        return None


def get_requests_session(access_token) -> Session:
    session = requests.Session()
    session.headers.update({"Content-type": "application/json", "Authorization": "Bearer " + str(access_token)})
    return session


def download_specs(access_token: str, environment_id: str):
    session = get_requests_session(access_token=access_token)

    manifest_url = urljoin(settings.MANIFEST_DEFAULT_HOST, environment_id)
    with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
        response = session.get(manifest_url)
    response.raise_for_status()

    manifest = response.json()
    return manifest


def download_manifest(deck: dict, authentication: TokenAuthentication, access_token: str, environment_index: int = 0):
    try:
        environment_id = deck["environment"][environment_index]["id"]
        console.info("Requesting manifests. This process may take a few seconds.")
        manifest = download_specs(
            access_token=access_token,
            environment_id=environment_id,
        )
    except HTTPError as e:
        project_id = deck["project"]["id"]
        if e.response.status_code == 404:
            console.warning(
                "This deck does potentially not specify a valid Environment of type 'local'. "
                f"Please go to https://app.unikube.io/project/{project_id}/decks "
                f"and save a valid values path."
            )
            exit(1)
        elif e.response.status_code == 403:
            console.warning("Refreshing access token")
            environment_id = deck["environment"][environment_index]["id"]
            response = authentication.refresh()
            if not response["success"]:
                console.exit_login_required()

            access_token = response["response"]["access_token"]
            try:
                manifest = download_specs(
                    access_token=access_token,
                    environment_id=environment_id,
                )
            except HTTPError as e:
                console.warning(f"Even after refreshing access token download specs fails with {e}")
                exit(1)
        else:
            console.error("Could not load manifest: " + str(e), _exit=True)

    return manifest


# environment
def environment_type_from_string(environment_type: str):
    try:
        environment_type = EnvironmentType(environment_type)
    except Exception as e:
        console.debug(e)
        environment_type = None

    return environment_type


def check_environment_type_local_or_exit(deck: dict, environment_index: int = 0):
    if (
        environment_type_from_string(environment_type=deck["environment"][environment_index]["type"])
        != EnvironmentType.LOCAL
    ):
        console.error("This deck cannot be installed locally.", _exit=True)


def check_running_cluster(ctx: ClickContext, cluster_provider_type: K8sProviderType.k3d, project_instance: dict):
    for cluster_data in ctx.cluster_manager.get_all():
        cluster = ctx.cluster_manager.select(cluster_data=cluster_data, cluster_provider_type=cluster_provider_type)
        if cluster.exists() and cluster.ready():
            if cluster.name == project_instance["title"] and cluster.id == project_instance["id"]:
                Telepresence(cluster.storage.get()).start()
                console.info(f"Kubernetes cluster for '{cluster.display_name}' is already running.", _exit=True)
            else:
                console.error(
                    f"You cannot start multiple projects at the same time. Project {cluster.name}({cluster.id}) is "
                    f"currently running. Please run 'unikube project down \"{cluster.name}({cluster.id})\"' first and "
                    f"try again.",
                    _exit=True,
                )



import os
import re
import signal
import sys
from time import sleep

import click
import click_spinner

from src import settings
from src.cli import console
from src.graphql import GraphQL
from src.local.providers.helper import get_cluster_or_exit
from src.local.system import Docker, KubeAPI, KubeCtl, Telepresence
from src.settings import UNIKUBE_FILE
from src.unikubefile.selector import unikube_file_selector


def get_deck_from_arguments(ctx, organization_id: str, project_id: str, deck_id: str):

    context = ctx.context.get(organization=organization_id, project=project_id, deck=deck_id)

    ## project_id
    cluster_list = ctx.cluster_manager.get_cluster_list(ready=True)
    cluster_choices = [f"{item.name} ({item.id})" for item in cluster_list]
    cluster_choices_ids = [item.id for item in cluster_list]

    # argument
    if not context.project_id:
        # argument from console
        project_selected = console.list(
            message="Please select a project",
            message_no_choices="No project is running.",
            choices=cluster_choices,
        )
        if project_selected is None:
            exit(1)

        project_id = re.search(r"\((.*?)\)", project_selected).group(1)
    else:
        project_id = context.project_id

    # check if project is in local storage
    if project_id not in cluster_choices_ids:
        console.error("The project cluster could not be found or you have another project activated.", _exit=True)

    cluster_data = ctx.cluster_manager.get(id=project_id)
    if not cluster_data:
        console.error("The cluster could not be found.", _exit=True)

    ## deck_id
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                allDecks(projectId: $id) {
                    results {
                        id
                        title
                        environment {
                            namespace
                        }
                    }
                }
            }
            """,
            query_variables={
                "id": project_id,
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    deck_list = data["allDecks"]["results"]
    deck_choices = [f'{item["title"]} ({item["id"]})' for item in deck_list]
    deck_choices_ids = [item["id"] for item in deck_list]

    # argument
    if not context.deck_id:
        # argument from console
        deck_selected = console.list(
            message="Please select a deck",
            message_no_choices="No deck found.",
            choices=deck_choices,
        )
        if deck_selected is None:
            exit(1)

        deck_id = re.search(r"\((.*?)\)", deck_selected).group(1)
    else:
        deck_id = context.deck_id

    if deck_id is None:
        console.exit_generic_error()

    # check if deck exists
    if deck_id not in deck_choices_ids:
        console.error("The deck could not be found.", _exit=True)

    # get deck
    deck = None
    for deck in deck_list:
        if deck["id"] == deck_id:
            break

    return cluster_data, deck


def argument_app(k8s, app: str):
    if not app:
        app_choices = [pod.metadata.name for pod in k8s.get_pods().items]
        app = console.list(
            message="Please select an app",
            choices=app_choices,
        )

    if not app:
        console.error("No apps available.", _exit=True)

    if app not in [pod.metadata.name for pod in k8s.get_pods().items]:
        console.error("App does not exist.", _exit=True)

    return app


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def list(ctx, organization, project, deck, **kwargs):
    """List all apps/pods."""

    ctx.auth.check()
    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # list
    k8s = KubeAPI(provider_data, deck)
    pod_table = [{"id": pod.metadata.uid, "name": pod.metadata.name} for pod in k8s.get_pods().items]

    console.table(data=pod_table)


@click.command()
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def info(ctx, organization, project, deck, **kwargs):
    raise NotImplementedError


@click.command()
def use(**kwargs):
    raise NotImplementedError


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_obj
def shell(ctx, app, organization=None, project=None, deck=None, **kwargs):
    """Drop into an interactive shell."""

    ctx.auth.check()
    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # shell
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    # get the data of the selected pod
    data = k8s.get_pod(app)

    # 1. check if this pod is of a switched deployment (in case of an active Telepresence)
    if data.metadata.labels.get("telepresence"):
        # the corresponding deployment by getting rid of the telepresence suffix
        deployment = "-".join(data.metadata.name.split("-")[0:-1])

        # the container name generated in "app switch" for that pod
        container_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
            project=cluster_data.name, deck=deck["title"], name=deployment
        ).replace(":", "")

        if Docker().check_running(container_name):
            # 2. Connect to that container
            # 2.a connect using Docker
            Docker().exec(container_name, "/bin/sh", interactive=True)
        else:
            console.error(
                "This is a Telepresence Pod with no corresponding Docker container "
                "running in order to connect (inconsistent state?)"
            )

    else:
        # 2.b connect using kubernetes
        KubeCtl(provider_data).exec_pod(app, deck["environment"][0]["namespace"], "/bin/sh", interactive=True)


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.pass_context
def exec(ctx, **kwargs):
    ctx.forward(shell)


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.option("--deployment", help="Specify the deployment if not set in the Unikubefile")
@click.option("--unikubefile", help="Specify the path to the Unikubefile", type=str)
@click.pass_obj
def switch(ctx, app, organization, project, deck, deployment, unikubefile, **kwargs):
    """Switch a running deployment with a local Docker image"""

    ctx.auth.check()
    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)

    # unikube file input
    if unikubefile:
        path_unikube_file = unikubefile
    else:
        path_unikube_file = os.path.join(os.getcwd(), UNIKUBE_FILE)

    unikube_file = unikube_file_selector.get(path_unikube_file=path_unikube_file)

    # 2: Get a deployment
    # 2.1.a Check the deployment identifier
    if not deployment and unikube_file:
        # 1.1.b check the unikubefile
        deployment = unikube_file.get_deployment()
        if not deployment:
            console.error("Please specify the 'deployment' ke of your app in your unikube.yaml.", _exit=True)
    else:
        console.error(
            "Please specify the deployment either using the '--deployment' option or in the unikube.yaml. "
            "Run 'unikube app switch' in a directory containing the unikube.yaml file.",
            _exit=True,
        )

    # 2.2 Fetch available "deployment:", deployments
    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query($id: UUID) {
                deck(id: $id) {
                    deployments(level: "local") {
                        id
                        title
                        description
                        ports
                        isSwitchable
                    }
                    environment {
                        id
                        type
                        valuesPath
                        namespace
                    }
                }
            }
            """,
            query_variables={
                "id": deck["id"],
            },
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    target_deployment = None

    for _deployment in data["deck"]["deployments"]:
        if _deployment["title"] == deployment:
            target_deployment = _deployment

    # 2.3 Check and select deployment data
    if target_deployment is None:
        console.error(
            f"The deployment '{deployment}' you specified could not be found.",
            _exit=True,
        )

    ports = target_deployment["ports"].split(",")
    deployment = target_deployment["title"]
    namespace = deck["environment"][0]["namespace"]

    console.info("Please wait while unikube prepares the switch.")
    with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
        # check telepresence
        provider_data = cluster.storage.get()
        telepresence = Telepresence(provider_data)

        available_deployments = telepresence.list(namespace, flat=True)
        if deployment not in available_deployments:
            console.error(
                "The given deployment cannot be switched. " f"You may have to run 'unikube deck install {deck}' first.",
                _exit=True,
            )

        is_swapped = telepresence.is_swapped(deployment, namespace)

    # 3: Build an new Docker image
    # 3.1 Grab the docker file
    context, dockerfile, target = unikube_file.get_docker_build()
    console.debug(f"{context}, {dockerfile}, {target}")
    console.info(f"Building a Docker image for {dockerfile} with context {context}")

    # 3.2 Set an image name
    image_name = settings.TELEPRESENCE_DOCKER_IMAGE_FORMAT.format(
        project=cluster_data.name.replace(" ", "").lower(), deck=deck["title"], name=deployment
    )

    docker = Docker()

    if is_swapped:
        console.warning("It seems this app is already switched in another process. ")
        if click.confirm("Do you want to kill it and switch here?"):
            telepresence.leave(deployment, namespace, silent=True)
            if docker.check_running(image_name):
                docker.kill(name=image_name)
        else:
            sys.exit(0)

    # 3.3 Build image

    with click_spinner.spinner(beep=False, disable=False, force=False, stream=sys.stdout):
        status, msg = docker.build(image_name, context, dockerfile, target)
    if not status:
        console.debug(msg)
        console.error("Failed to build Docker image.", _exit=True)

    console.info(f"Docker image successfully built: {image_name}")

    # 4. Start the Telepresence session
    # 4.1 See if there are volume mounts
    mounts = unikube_file.get_mounts()
    console.debug(f"Volumes requested: {mounts}")

    # 4.2 See if there special env variables
    envs = unikube_file.get_environment()
    console.debug(f"Envs requested: {envs}")

    # 4.3 See if there is a run command to be executed
    command = unikube_file.get_command(port=ports[0])
    console.debug(f"Run command: {command}")

    console.info("Starting your container, this may take a while to become effective")

    telepresence.swap(deployment, image_name, command, namespace, envs, mounts, ports[0])
    if docker.check_running(image_name):
        docker.kill(name=image_name)


@click.command()
def pulldb(**kwargs):
    raise NotImplementedError


@click.command()
@click.argument("app", required=False)
@click.option("--organization", "-o", help="Select an organization")
@click.option("--project", "-p", help="Select a project")
@click.option("--deck", "-d", help="Select a deck")
@click.option("--follow", "-f", is_flag=True, default=False, help="Follow logs.")
@click.pass_obj
def logs(ctx, app, organization=None, project=None, deck=None, follow=False, **kwargs):
    """Display the container's logs"""

    ctx.auth.check()
    cluster_data, deck = get_deck_from_arguments(ctx, organization, project, deck)

    # get cluster
    cluster = get_cluster_or_exit(ctx, cluster_data.id)
    provider_data = cluster.storage.get()

    # log
    k8s = KubeAPI(provider_data, deck)
    app = argument_app(k8s, app)

    logs = k8s.get_logs(app, follow)

    # output
    click.echo(logs)


@click.command()
def expose(**kwargs):
    raise NotImplementedError


@click.command()
def env(**kwargs):
    raise NotImplementedError


@click.command()
def request_env(**kwargs):
    raise NotImplementedError



from click.testing import CliRunner

from src.cli import deck
from unikube import ClickContext


def check():
    """Function used to mock check function"""
    pass


def test_list():
    runner = CliRunner()
    obj = ClickContext()
    obj.auth.check = check
    result = runner.invoke(
        deck.list,
        [
            "--organization",
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
            "--project",
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
        ],
        obj=obj,
    )
    assert result.exit_code == 1


def test_info():
    runner = CliRunner()
    obj = ClickContext()
    obj.auth.check = check
    result = runner.invoke(
        deck.info,
        [
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
        ],
        obj=obj,
    )
    assert result.exit_code == 1


def test_ingress():
    runner = CliRunner()
    obj = ClickContext()
    obj.auth.check = check
    result = runner.invoke(
        deck.ingress,
        [
            "13fc0b1b-3bc1-4a69-8e80-835fb1515bc4",
        ],
        obj=obj,
    )
    assert result.exit_code == 1



import pytest
from requests import HTTPError, Session

from src.helpers import (
    download_manifest,
    download_specs,
    get_requests_session,
    select_entity,
    select_entity_from_cluster_list,
)
from src.local.providers.types import K8sProviderData


def test_get_requests_session():
    access_token = ""
    assert type(access_token) is str

    session = get_requests_session(access_token=access_token)
    assert type(session) is Session


def test_download_specs():
    environment_id = "WRONG"
    access_token = ""

    with pytest.raises(HTTPError) as pytest_wrapped_e:
        _ = download_specs(access_token=access_token, environment_id=environment_id)

    assert pytest_wrapped_e.type == HTTPError


def test_download_manifest():
    deck = {
        "environment": [
            {
                "id": None,
            }
        ],
        "project": {
            "id": None,
        },
    }

    class Authentication:
        def refresh(self):
            return {"success": True, "response": {"access_token": ""}}

    access_token = ""
    authentication = Authentication()

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        _ = download_manifest(deck=deck, authentication=authentication, access_token=access_token)

    assert pytest_wrapped_e.type == SystemExit
    assert pytest_wrapped_e.value.code == 1


def test_select_entity():
    entity_list = [{"id": "random-id-1-2-3", "title": "test-select-entity", "clusterSettings": {"id": "1", "port": 0}}]
    identifier = "test-select-entity(random-id-1-2-3)"
    response = select_entity(entity_list, identifier)

    assert response == entity_list[0]


def test_select_entity_one_similar_entity():
    entity_list = [{"id": "random-id-1-2-3", "title": "test-select-entity", "clusterSettings": {"id": "1", "port": 0}}]
    identifier = "test-select-entity"
    response = select_entity(entity_list, identifier)
    assert response is entity_list[0]


def test_select_entity_duplicate_entities():
    entity_list = [
        {"id": "random-id-1-2-3", "title": "test-select-entity", "clusterSettings": {"id": "1", "port": 0}},
        {"id": "random-id-3-2-1", "title": "test-select-entity", "clusterSettings": {"id": "1", "port": 0}},
    ]
    identifier = "test-select-entity"
    response = select_entity(entity_list, identifier)
    assert response is None


def test_select_entity_from_cluster_list():
    entity_list = [K8sProviderData(id="random-id-1-2-3", name="test-select-entity")]
    identifier = "test-select-entity(random-id-1-2-3)"
    response = select_entity_from_cluster_list(entity_list, identifier)

    assert response == entity_list[0]


def test_select_entity_from_cluster_list_one_similar_entity():
    entity_list = [K8sProviderData(id="random-id-1-2-3", name="test-select-entity")]
    identifier = "test-select-entity"
    response = select_entity_from_cluster_list(entity_list, identifier)
    assert response is entity_list[0]


def test_select_entity_from_cluster_list_duplicate_entities():
    entity_list = [
        K8sProviderData(id="random-id-1-2-3", name="test-select-entity"),
        K8sProviderData(id="random-id-3-2-1", name="test-select-entity"),
    ]
    identifier = "test-select-entity"
    response = select_entity_from_cluster_list(entity_list, identifier)
    assert response is None



import os
from typing import Union

import yaml

from src.unikubefile.unikube_file import UnikubeFile, UnikubeFileVersionError
from src.unikubefile.unikube_file_1_0 import UnikubeFile_1_0


class UnikubeFileSelector:
    def __init__(self, options: dict):
        self.options = options

    def get(self, path_unikube_file: str) -> Union[UnikubeFile, UnikubeFile_1_0, None]:
        # check if file exists
        if not os.path.isfile(path_unikube_file):
            return None

        # load unikube file + get version
        with open(path_unikube_file) as unikube_file:
            data = yaml.load(unikube_file, Loader=yaml.FullLoader)
            version = data.get("version", "latest")

        # get class
        unikube_file_class = self.options.get(version)
        if not unikube_file_class:
            raise UnikubeFileVersionError

        return unikube_file_class(
            path=path_unikube_file,
            data=data,
        )


unikube_file_selector = UnikubeFileSelector(
    options={
        "latest": UnikubeFile_1_0,
        "1": UnikubeFile_1_0,
        "1.0": UnikubeFile_1_0,
    }
)



from abc import ABC, abstractmethod

from src.local.providers.types import K8sProviderData
from src.local.system import Docker


class IK8sProviderStorage(ABC):
    @abstractmethod
    def get(self) -> K8sProviderData:
        raise NotImplementedError

    @abstractmethod
    def set(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> None:
        raise NotImplementedError


class AbstractK8SProviderStorage(IK8sProviderStorage):
    def __init__(
        self,
        id: str,
        storage=None,
    ) -> None:
        super().__init__()

        self.id = id
        self.storage = storage

    def get(self) -> K8sProviderData:
        return self.storage.get(self.id)

    def set(self, data) -> None:
        self.storage.set(self.id, data)

    def delete(self) -> None:
        self.storage.delete(self.id)


class IK8sProvider(ABC):
    @abstractmethod
    def create(self, ingress_port: int = None) -> bool:
        raise NotImplementedError

    @abstractmethod
    def start(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def stop(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def delete(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def exists(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def ready(self) -> bool:
        raise NotImplementedError


class AbstractK8sProvider(IK8sProvider):
    provider_type = None

    def __init__(
        self,
        id: str,
        name: str = None,
        storage: AbstractK8SProviderStorage = None,
    ) -> None:
        self.id = id
        self.name = name
        self.storage = storage

    @property
    def display_name(self):
        name = self.name
        if name:
            return name

        id = self.id
        return id

    @property
    def k8s_provider_type(self):
        return self.provider_type

    def ready(self) -> bool:
        # get name
        provider_data = self.storage.get()
        name = provider_data.name
        if not name:
            return False

        return Docker().check_running(name)



from urllib.parse import urljoin

import jwt
import requests

import src.cli.console as console
from src import settings
from src.authentication.types import AuthenticationData
from src.storage.general import LocalStorageGeneral
from src.storage.user import LocalStorageUser


class IAuthentication:
    def check(self):
        raise NotImplementedError

    def login(self, email: str, password: str) -> dict:
        raise NotImplementedError

    def logout(self):
        raise NotImplementedError

    def verify(self) -> dict:
        raise NotImplementedError

    def refresh(self) -> dict:
        raise NotImplementedError

    def verify_or_refresh(self) -> bool:
        raise NotImplementedError


class TokenAuthentication(IAuthentication):
    def __init__(
        self,
        local_storage_general: LocalStorageGeneral,
        timeout=settings.TOKEN_TIMEOUT,
    ):
        self.local_storage_general = local_storage_general

        self.general_data = self.local_storage_general.get()
        self.timeout = timeout

        self.url_public_key = urljoin(self.__get_host(), settings.TOKEN_PUBLIC_KEY)
        self.url_login = urljoin(self.__get_host(), settings.TOKEN_LOGIN_PATH)
        self.url_verify = urljoin(self.__get_host(), settings.TOKEN_VERIFY_PATH)
        self.url_refresh = urljoin(self.__get_host(), settings.TOKEN_REFRESH_PATH)

        self.client_id = settings.KC_CLIENT_ID

        # RPT
        self.requesting_party_token_audience = settings.TOKEN_RPT_AUDIENCE

    def __get_host(self) -> str:
        try:
            local_storage_user = LocalStorageUser(user_email=self.general_data.authentication.email)
            user_data = local_storage_user.get()
            auth_host = user_data.config.auth_host

            if not auth_host:
                raise Exception("User data config does not specify an authentication host.")

            return auth_host

        except Exception:
            return settings.AUTH_DEFAULT_HOST

    def _get_requesting_party_token(self, access_token):
        # requesting party token (RPT)
        response = self.__request(
            url=self.url_login,
            data={
                "audience": self.requesting_party_token_audience,
                "grant_type": "urn:ietf:params:oauth:grant-type:uma-ticket",
            },
            headers={"Authorization": f"Bearer {access_token}"},
            message_exception="Could not establish a server connection.",
            message_200="",
            message_400="Wrong user credentials or account does not exist.",
            message_500="There was an server error.",
        )

        # select response
        if not response["success"]:
            return response

        return response

    def check(self):
        # login required
        if not self.verify_or_refresh():
            console.exit_login_required()

    def login(
        self,
        email: str,
        password: str,
    ) -> dict:
        # set/update user config
        local_storage_user = LocalStorageUser(user_email=email)
        user_data = local_storage_user.get()
        user_data.config.auth_host = settings.AUTH_DEFAULT_HOST
        local_storage_user.set(user_data)

        # access token + refresh token
        response_token = self.__request(
            url=self.url_login,
            data={
                "username": email,
                "password": password,
                "grant_type": "password",
                "client_id": self.client_id,
            },
            message_exception="Could not establish a server connection.",
            message_200="",
            message_400="Wrong user credentials or account does not exist.",
            message_500="There was an server error.",
        )

        if not response_token["success"]:
            return response_token

        # requesting party token (RPT)
        response_RPT = self._get_requesting_party_token(response_token["response"]["access_token"])

        # select response
        if response_RPT["success"]:
            response = response_RPT
            requesting_party_token = True
        else:
            response = response_token
            requesting_party_token = False

        # set authentication data
        self.general_data.authentication = AuthenticationData(
            email=email,
            access_token=response["response"]["access_token"],
            refresh_token=response["response"]["refresh_token"],
            requesting_party_token=requesting_party_token,
        )

        self.local_storage_general.set(self.general_data)

        return response

    def logout(self):
        self.general_data.authentication = AuthenticationData()
        self.local_storage_general.set(self.general_data)

    def verify(self) -> dict:
        # keycloak
        access_token = self.general_data.authentication.access_token
        response = self.__request(
            url=self.url_verify,
            data={},
            headers={"Authorization": f"Bearer {access_token}"},
            message_exception="Could not establish a server connection.",
            message_200="",
            message_400="Invalid or expired login data, please log in again with 'unikube login'.",
            message_500="There was an server error.",
        )
        return response

    def refresh(self) -> dict:
        # request
        refresh_token = self.general_data.authentication.refresh_token
        response_token = self.__request(
            url=self.url_refresh,
            data={
                "refresh_token": refresh_token,
                "grant_type": "refresh_token",
                "client_id": self.client_id,
            },
            message_exception="Could not establish a server connection.",
            message_200="",
            message_400="Refresh token expired or account does not exist.",
            message_500="There was an server error.",
        )

        if not response_token["success"]:
            return response_token

        # requesting party token (RPT)
        response_RPT = self._get_requesting_party_token(response_token["response"]["access_token"])

        # select response
        if response_RPT["success"]:
            response = response_RPT
            requesting_party_token = True
        else:
            response = response_token
            requesting_party_token = False

        # update token
        if response["success"]:
            self.general_data = self.local_storage_general.get()
            self.general_data.authentication.access_token = response["response"]["access_token"]
            self.general_data.authentication.refresh_token = response["response"]["refresh_token"]
            self.general_data.authentication.requesting_party_token = requesting_party_token
            self.local_storage_general.set(self.general_data)

        return response

    def verify_or_refresh(self) -> bool:
        # verify
        response = self.verify()
        if response["success"]:
            return True

        # refresh
        response = self.refresh()
        if response["success"]:
            return True

        # exception messsage
        console.debug(response["message"])

        return False

    def __request(
        self,
        url,
        data,
        message_exception,
        message_200,
        message_400,
        message_500,
        headers=None,
    ) -> dict:
        # request
        try:
            req = requests.post(
                url,
                data,
                headers=headers,
                timeout=self.timeout,
            )
        except Exception as e:
            console.debug(e)
            return {
                "success": False,
                "message": message_exception,
                "response": None,
            }

        # return
        if req.status_code == 200:
            success = True
            message = message_200

        elif req.status_code in [400, 401, 404]:
            success = False
            message = message_400

        elif req.status_code in [500, 501, 502, 503]:
            success = False
            message = message_500

        else:
            success = False
            message = ""

        # get json response
        try:
            response = req.json()
        except Exception:
            response = None

        return {
            "success": success,
            "message": message,
            "response": response,
        }

    def token_from_response(self, response):
        token = jwt.decode(
            response["response"]["access_token"],
            algorithms=["RS256"],
            audience=settings.TOKEN_AUDIENCE,
            options={"verify_signature": False},
        )
        return token


def get_authentication():
    token_authentication = TokenAuthentication(local_storage_general=LocalStorageGeneral())
    return token_authentication



# -*- coding: utf-8 -*-
import os
from typing import List, Optional, Tuple

from src.context.types import ContextData
from src.unikubefile.unikube_file import UnikubeFile, UnikubeFileError


class UnikubeFile_1_0(UnikubeFile):
    def __init__(self, path: str, data: dict):
        super().__init__(path=path, data=data)
        self.path = path

        self._data = data
        self._verify()

        # set the first app active for all query methods
        self._app = self.get_apps()[0]
        self._path = os.path.dirname(path)

    def get_context(self):
        context = self._data.get("context", None)
        if not context:
            return ContextData()

        return ContextData(
            organization=context.get("organization", None),
            project=context.get("project", None),
            deck=context.get("deck", None),
        )

    def _verify(self):
        self._check_required("apps", self._data, "root")

    def _check_required(self, _keys, node, node_title):
        if type(_keys) == list:
            for _key in _keys:
                if _key not in node:
                    raise UnikubeFileError(f"Unikubefile not valid, missing key {_key} not in {node_title}")
        else:
            if _keys not in node:
                raise UnikubeFileError(f"Unikubefile not valid, missing key {_keys} not in {node_title}")

    # def set_app(self, app):
    #     if app not in self.get_apps():
    #         raise UnikubeFileError(f"the app '{app}' is not part of this Unikubefile")

    def get_apps(self):
        apps = list(self._data["apps"].keys())
        return apps

    def get_docker_build(self) -> Tuple[str, str, str]:
        app = self._data["apps"][self._app]
        build = app.get("build")
        if build:
            base_path = os.path.dirname(self.path)
            path = os.path.abspath(os.path.join(base_path, build["context"]))
            dockerfile = os.path.join(base_path, build.get("dockerfile", "Dockerfile"))
            return path, dockerfile, build.get("target", "")
        else:
            return os.path.abspath("."), "Dockerfile", ""

    def get_command(self, **format) -> Optional[str]:
        app = self._data["apps"][self._app]
        command = app.get("command")
        if command:
            command = command.format(**format)
            return command.split(" ")
        else:
            return None

    def get_deployment(self) -> Optional[str]:
        app = self._data["apps"][self._app]
        deployment = app.get("deployment")
        if deployment:
            return str(deployment)
        else:
            return None

    def get_mounts(self) -> List[Tuple[str, str]]:
        app = self._data["apps"][self._app]
        volumes = app.get("volumes")
        mounts = []
        if volumes:
            base_path = os.path.dirname(self.path)
            for mount in volumes:
                mount = mount.split(":")
                source = os.path.abspath(os.path.join(base_path, mount[0]))
                target = mount[1]
                mounts.append((source, target))
        return mounts

    def get_environment(self) -> List[Tuple[str, str]]:
        app = self._data["apps"][self._app]
        variables = app.get("environment")
        if variables is None:
            # fallback to the env keyword
            variables = app.get("env")
        envs = []
        if variables:
            for env in variables:
                envs.append(list(env.items())[0])
        return envs



from enum import Enum

import click

from src import settings


# log level
class LogLevel(str, Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    SUCCESS = "success"


log_level_mapping = {
    LogLevel.DEBUG: [LogLevel.DEBUG],
    LogLevel.INFO: [LogLevel.DEBUG, LogLevel.INFO],
    LogLevel.WARNING: [LogLevel.DEBUG, LogLevel.INFO, LogLevel.WARNING],
    LogLevel.ERROR: None,  # None -> print always
    LogLevel.SUCCESS: None,
}

# color
color_mapping = {
    LogLevel.DEBUG: "cyan",
    LogLevel.INFO: "",
    LogLevel.WARNING: "yellow",
    LogLevel.ERROR: "red",
    LogLevel.SUCCESS: "green",
}


# helper
def _click_secho(msg: str, silent: bool, log_level: str = None, _exit: bool = False, _exit_code: int = 1, **kwargs):
    # get log level settings and console mapping
    setting_log_level = LogLevel(settings.CLI_LOG_LEVEL.lower())
    console_log_level = log_level_mapping.get(log_level, None)

    # check log level
    if console_log_level is None:
        pass  # log level independent

    else:
        if setting_log_level not in console_log_level:
            return None

    # silence message?
    if silent:
        return None

    # color
    fg = color_mapping.get(log_level, "")

    # console echo
    click.secho(f"[{log_level.value.upper()}] {msg}", fg=fg, **kwargs)

    # exit
    if _exit:
        exit(_exit_code)


# console output
def debug(msg: str, silent: bool = False, **kwargs):
    _click_secho(msg, silent, log_level=LogLevel.DEBUG, **kwargs)


def info(msg: str, silent: bool = False, _exit: bool = False, _exit_code: int = 1, **kwargs):
    _click_secho(msg=msg, silent=silent, log_level=LogLevel.INFO, _exit=_exit, _exit_code=_exit_code, **kwargs)


def warning(msg: str, silent: bool = False, _exit: bool = False, _exit_code: int = 1, **kwargs):
    _click_secho(msg=msg, silent=silent, log_level=LogLevel.WARNING, _exit=_exit, _exit_code=_exit_code, **kwargs)


def error(msg: str, _exit: bool = False, _exit_code: int = 1, **kwargs):
    _click_secho(msg=msg, silent=False, log_level=LogLevel.ERROR, _exit=_exit, _exit_code=_exit_code, **kwargs)


def success(msg: str, silent: bool = False, **kwargs):
    _click_secho(msg=msg, silent=silent, log_level=LogLevel.SUCCESS, **kwargs)



from pydantic import BaseModel


class AuthenticationData(BaseModel):
    email: str = ""
    access_token: str = ""
    refresh_token: str = ""
    requesting_party_token: bool = False
    public_key: str = ""



import sys

import click

import src.cli.console as console
from src.cli import app as app_cmd
from src.cli import auth as auth_cmd
from src.cli import deck as deck_cmd
from src.cli import orga as orga_cmd
from src.cli import project as project_cmd
from src.cli import system as system_cmd
from src.context import ClickContext

version = sys.version_info
if version.major == 2:
    console.error("Python 2 is not supported for unikube. Please upgrade python.")
    exit(1)


# click -----
@click.group()
@click.version_option()
@click.pass_context
def cli(ctx, **kwargs):
    ctx.obj = ClickContext()


# system
@cli.group()
@click.pass_obj
def system(ctx):
    """
    Manage dependencies on your local machine. Install and verify required tools to get your cluster running.
    """


system.add_command(system_cmd.install)
system.add_command(system_cmd.verify)


# organization
@cli.group()
@click.pass_obj
def orga(ctx):
    """
    Manage your organizations.
    """


orga.add_command(orga_cmd.list)
orga.add_command(orga_cmd.info)
orga.add_command(orga_cmd.use)


# project
@cli.group()
@click.pass_obj
def project(ctx):
    """
    Manage your projects.
    """


project.add_command(project_cmd.list)
project.add_command(project_cmd.info)
project.add_command(project_cmd.use)
project.add_command(project_cmd.up)
project.add_command(project_cmd.down)
project.add_command(project_cmd.delete)


# deck
@cli.group()
@click.pass_obj
def deck(ctx):
    """
    Manage your decks.
    """


deck.add_command(deck_cmd.list)
deck.add_command(deck_cmd.info)
deck.add_command(deck_cmd.use)
deck.add_command(deck_cmd.install)
deck.add_command(deck_cmd.uninstall)
deck.add_command(deck_cmd.ingress)


# application
@cli.group()
@click.pass_obj
def app(ctx):
    """
    Manage your applications.
    """


app.add_command(app_cmd.info)
app.add_command(app_cmd.list)
app.add_command(app_cmd.use)
app.add_command(app_cmd.shell)
app.add_command(app_cmd.switch)
app.add_command(app_cmd.pulldb)
app.add_command(app_cmd.logs)
app.add_command(app_cmd.expose)
app.add_command(app_cmd.env)
app.add_command(app_cmd.request_env)
app.add_command(app_cmd.exec)


# authentication
@cli.group()
def auth():
    """
    Manage Unikube's authentication state.
    """


auth.add_command(auth_cmd.login)
auth.add_command(auth_cmd.logout)
auth.add_command(auth_cmd.status)


# shortcut
# -> include auth check in functions if required
cli.add_command(auth_cmd.login)
cli.add_command(auth_cmd.logout)
cli.add_command(project_cmd.up)
cli.add_command(deck_cmd.install)
cli.add_command(app_cmd.shell)


if __name__ == "__main__":
    cli()



import click

import src.cli.console as console
from src.graphql import GraphQL
from src.helpers import select_entity
from src.keycloak.permissions import KeycloakPermissions
from src.storage.user import get_local_storage_user


@click.command()
@click.pass_obj
def list(ctx, **kwargs):
    """
    List all your organizations.
    """

    context = ctx.context.get()

    # keycloak
    try:
        keycloak_permissions = KeycloakPermissions(authentication=ctx.auth)
        permission_list = keycloak_permissions.get_permissions_by_scope("organization:*")
    except Exception as e:
        permission_list = None
        console.debug(e)
        console.exit_generic_error()

    # append "(active)"
    if context.organization_id:
        for permission in permission_list:
            if permission.rsid == context.organization_id:
                permission.rsid += " (active)"

    # console
    organization_list = [
        {
            "id": permission.rsid,
            "name": permission.rsname.replace("organization ", ""),
        }
        for permission in permission_list
    ]
    console.table(
        data=organization_list,
        headers={
            "id": "id",
            "name": "name",
        },
    )


@click.command()
@click.argument("organization", required=False)
@click.pass_obj
def info(ctx, organization, **kwargs):
    """
    Display further information of the selected organization.
    """

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            {
                allOrganizations {
                    results {
                        id
                        title
                        description
                    }
                }
            }
            """
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    organization_list = data["allOrganizations"]["results"]

    # argument
    if not organization:
        # argument from context
        context = ctx.context.get()
        if context.organization_id:
            organization_instance = ctx.context.get_organization()
            organization = organization_instance["title"] + f"({organization_instance['id']})"

        # argument from console
        else:
            organization = console.list(
                message="Please select an organization",
                choices=[organization["title"] + f"({organization['id']})" for organization in organization_list],
            )
            if organization is None:
                return None

    # select
    organization_selected = select_entity(organization_list, organization)

    # console
    if organization_selected:
        console.table(
            data={
                "key": [k for k in organization_selected.keys()],
                "value": [v for v in organization_selected.values()],
            },
            headers=["Key", "Value"],
        )
    else:
        console.error("Organization does not exist.")


@click.command()
@click.argument("organization_id", required=False)
@click.option("--remove", "-r", is_flag=True, default=False, help="Remove local organization context")
@click.pass_obj
def use(ctx, organization_id, remove, **kwargs):
    """
    Set local organization context.
    """

    # user_data / context
    local_storage_user = get_local_storage_user()
    user_data = local_storage_user.get()

    # option: --remove
    if remove:
        user_data.context.deck_id = None
        user_data.context.project_id = None
        user_data.context.organization_id = None
        local_storage_user.set(user_data)
        console.success("Organization context removed.")
        return None

    # GraphQL
    try:
        graph_ql = GraphQL(authentication=ctx.auth)
        data = graph_ql.query(
            """
            query {
                allOrganizations {
                    results {
                        id
                        title
                    }
                }
            }
            """
        )
    except Exception as e:
        data = None
        console.debug(e)
        console.exit_generic_error()

    organization_list = data["allOrganizations"]["results"]
    organization_dict = {organization["id"]: organization["title"] for organization in organization_list}

    # argument
    if not organization_id:
        organization_title = console.list(
            message="Please select an organization",
            choices=organization_dict.values(),
        )
        if organization_title is None:
            return False

        for id, title in organization_dict.items():
            if title == organization_title:
                organization_id = id

    # set organization
    user_data.context.deck_id = None
    user_data.context.project_id = None
    user_data.context.organization_id = organization_id
    local_storage_user.set(user_data)

    console.success(f"Organization context: {user_data.context}")



from click.testing import CliRunner

from src.cli import orga
from unikube import ClickContext


def test_login():
    runner = CliRunner()
    result = runner.invoke(
        orga.list,
        obj=ClickContext(),
    )
    assert "[DEBUG] Refresh token expired or account does not exist." in result.output
    assert "[INFO] You need to login (again). Please run 'unikube login' and try again." in result.output
